﻿-- Author      : RisM
-- Create Date : 6/28/2009 3:59:16 PM

local SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)
local DEFAULT_SPEECHES = LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_DEFAULT_SPEECHES",	false)

SpeakinSpell:PrintLoading("oldversions.lua")

-------------------------------------------------------------------------------
-- UPDATE OLD SAVED DATA
-------------------------------------------------------------------------------


--[[
SpeakinSpell:ConvertOldSpellTableToNewEventTable()

-----------------------------------------------------
-- OLD FORMAT: 3.1.2.03
local spellnamekey = name
if rank then
	spellnamekey = name.."("..rank..")"
end
-- rank can also be "_ event: Login" or "_ macro something"

SpeakinSpellSavedData.EventTable[spellnamekey] = {
	-- default options
	Channels = self:CopyTable(DEFAULT_CHANNEL_RULES),
	WhisperTarget = false,
	Frequency = 1.00, --default 100%
	Messages = self:CopyTable(DEFAULT_SPEECHES.NEWSPELL),
}

-----------------------------------------------------
-- NEW FORMAT: 3.1.2.05
-- DetectedEvent table has been added
-- key has changed
-- new key format is name..rank..type

SpeakinSpellSavedData.EventTable[de.key] = {
	-- event information
	DetectedEvent = {
		key			= de.key,
		name		= de.name,
		rank		= de.rank,
		type		= de.type,
		--DisplayName = de.DisplayName, -- no longer stored in data as of 3.2.2.02
		--SpellLink	= de.SpellLink, --no longer stored in data as of 3.2.0.08
		-- caster	= de.caster -- not stored in event table
		-- target	= de.target -- not stored in event table
		-- pet		= de.pet    -- not stored in event table
	},
	-- default options
	Channels = self:CopyTable(DEFAULT_CHANNEL_RULES),
	WhisperTarget = false,
	Frequency = 1.00, --default 100%
	Cooldown = 0,
	Messages = self:CopyTable(DEFAULT_SPEECHES.NEWSPELL),
}
--]]

-- oldkey is a v3.1.2.03 SpellTable key
-- returns a new DetectedEventStub object for 3.1.2.05
function SpeakinSpell:FixOld_31205_CreateDEStub(oldkey)

	local stub = {
		name = "",
		rank = "",
		type = "", --unknown
		-- in 3.1.2.03, the only detected events were self cast on player
		-- or equivalent for as much as we cared at the time.
		caster = UnitName("player"),
		target = UnitName("player"),
	}
	
	-- check for macros using the 3.1.2.03 method of key names
	if self:StartsWith(oldkey, "_ macro") or self:StartsWith(oldkey, "_ "..L["/macro"]) then
		--remove leading "_ " from the name
		stub.name = string.sub(oldkey, 3)
		stub.rank = ""
		stub.type = "MACRO"
		return stub
	end
	
	-- check for the only event that was supported in 3.1.2.03
	if oldkey == "_ event: Login" then
		stub.name = L.EVENTNAMES["SpeakinSpell Loaded"]
		stub.rank = ""
		stub.type = "EVENT"
		return stub
	end
	
	-- it must be either a UNIT_SPELLCAST_SENT or a SPELL_AURA_APPLIED_BUFF
	-- let's take a closer look... 
	
	-- GetSpellInfo
	local name, rank, icon, cost, isFunnel, powerType, castTime, minRange, maxRange = GetSpellInfo(oldkey)

	-- check whether they old key specified the rank or not
	local oldkeyrank = string.match(oldkey, "%b()")
	
	-- if GetSpellInfo returns data, then it's a real spell, which will mean UNIT_SPELLCAST_SENT
	if name then
		stub.type = "UNIT_SPELLCAST_SENT"
		stub.name = name
		if oldkeyrank and rank then
			stub.rank = rank
		else -- use unspecified rank
			stub.rank = ""
		end
		return stub
	end

	-- else, do some manual parsing and see what we can come up with
	
	-- parse the old key name for the name and rank
	stub.name = oldkey
	stub.rank = string.match(oldkey, "%b()")
	
	if not stub.rank then
		stub.rank = ""
	end
	
	if stub.rank ~= "" then
		stub.name = string.sub(oldkey, 1, -2 - string.len(stub.rank) ) -- remove rank from name
		stub.rank = string.sub(stub.rank, 2, -2) --remove wrapping parens
	end
	-- else the name is the full key, already set up above
	
	stub.name = self:TitleCase(stub.name)
	stub.rank = self:TitleCase(stub.rank)
	
	-- buffs received have no rank
	-- if this has rank, it's definitely not a buff
	if stub.rank ~= "" then
		stub.type = "UNIT_SPELLCAST_SENT"
		return stub
	end
	
--	self:Print("---------")
--	self:Print("GetSpellInfo - assumed buff")
--	self:Print("name:"..tostring(name))
--	self:Print("rank:"..tostring(rank))
--	self:Print("icon:"..tostring(icon))
--	self:Print("cost:"..tostring(cost))
--	self:Print("isFunnel:"..tostring(isFunnel))
--	self:Print("powerType:"..tostring(powerType))
--	self:Print("castTime:"..tostring(castTime))
--	self:Print("minRange:"..tostring(minRange))
--	self:Print("maxRange:"..tostring(maxRange))
--	self:Print("---------")

	-- can't tell for sure, so use both
	stub.type = "BOTH"
	
	return stub
end



function SpeakinSpell:FixOld_31205_CreateDetectedEvent( DetectedEventStub )

	local de = self:CopyTable(DetectedEventStub)
	
	-- SpellLink --no longer stored in data as of 3.2.0.08 - safe to remove from legacy support
	-- create a spell link, if possible
	-- the stub will set SpellLink = true if the event is a spell
	-- else will we make a fake one
--	if de.SpellLink then
--		de.SpellLink,_ = GetSpellLink(de.name, de.rank)
--	end
	-- NOTE: some aura gains are not linkable, notably "Refreshment" from "Drink"
	--		in that case, GetSpellLink will return NULL
--	if not de.SpellLink then
--		de.SpellLink = "["..de.name.."]"
--	end
	
	--[[ don't need this for importing saved data
	
	-- validate the target - get a backup target from selection, focus, or assumed self-cast
	if not de.target or de.target == "" or de.target == "Unknown" then
		de.target = self:GetDefaultTarget(true)
	end
	-- END SpellLink --no longer stored in data as of 3.2.0.08

	-- fetch additional target information
	de.targetclass = UnitClass(de.target)
	de.targetrace = UnitRace(de.target)

	--setup <pet> (3.2.0.06 no longer created here)
	de.pet = UnitName("pet")
	--]]
	
	-- create a display name for this event
	-- no longer stored in event table as of 32202
--	if de.rank ~= "" then
--		de.DisplayName = L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name.." ("..de.rank..")" 
--	else
--		de.DisplayName = L.EVENTTYPES.IN_SPELL_LIST[de.type]..de.name
--	end
	
	-- create a unique key for this event
	de.key = tostring(de.name)..tostring(de.rank)..tostring(de.type) --DO NOT CHANGE 3.1.2.05 format is intentional
	
	--[[ don't need this for importing saved data
	
	-- check for a EventTableEntry for this specific rank
	de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
	
	-- if that wasn't found, check for a rankless entry
	if not de.EventTableEntry then
		local ranklesskey = de.name..de.type
		de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
		-- if that was found, then update the key for future reference
		if de.EventTableEntry then
			de.key = ranklesskey
		end
	end
	--]]
	
	return de
end


function SpeakinSpell:FixOld_31205_CreateEventTableEntry(oldkey, OldSpellTableEntry, de)
	SpeakinSpellSavedData.EventTable[de.key] = {
		-- event information
		DetectedEvent = {
			key			= de.key,
			name		= de.name,
			rank		= de.rank,
			type		= de.type,
			--DisplayName = de.DisplayName, -- no longer stored in event table as of 32202
			--SpellLink	= de.SpellLink, --no longer stored in event table as of 32008
			-- caster	= de.caster -- not stored in event table in 31205
			-- target	= de.target -- not stored in event table in 31205
			-- pet		= de.pet    -- not stored in event table in 31205
		},
		-- default options
		-- Channels and Messages tables are guaranteed to exist
		-- the rest might not - correct for that below
		Channels		= self:CopyTable(OldSpellTableEntry.Channels),
		Messages		= self:CopyTable(OldSpellTableEntry.Messages),
	}
	
	-- validate missing data and set defaults
	-- this used to be performed as a side effect in the GUI prior to 3.1.2.05
	local de = SpeakinSpellSavedData.EventTable[de.key]
	
	--WhisperTarget
	if OldSpellTableEntry.WhisperTarget then
		de.WhisperTarget = OldSpellTableEntry.WhisperTarget
	else
		de.WhisperTarget = false
	end
	
	--Frequency
	if OldSpellTableEntry.Frequency then
		de.Frequency = OldSpellTableEntry.Frequency
	else
		de.Frequency = 1.00 --FREQUENCY_DEFAULT
	end
	
	--Cooldown
	if OldSpellTableEntry.Cooldown then
		de.Cooldown	= OldSpellTableEntry.Cooldown
	else
		de.Cooldown	= 0 --COOLDOWN_DEFAULT
	end
	
	--DisableAnnouncements
	if OldSpellTableEntry.DisableAnnouncements then
		de.DisableAnnouncements	= OldSpellTableEntry.DisableAnnouncements
	else
		de.DisableAnnouncements	= false
	end
end


function SpeakinSpell:FixOld_31205_ImportSpellTable()

	SpeakinSpellSavedData.EventTable = {}

	for oldkey,OldSpellTableEntry in pairs(SpeakinSpellSavedData.SpellTable) do
	
		-- the process in 31205 is basically 3 steps
		-- 1) declare a DetectedObjectStub with name, rank, type, caster, and target
		--		we will create this by parsing the oldkey value, which is where this info lived
		-- 2) pass that stub through the function FullDetectedEvent = CreateDetectedEvent(stub)
		--		we have a copy of that function, with some small tweaks and simplifications
		-- 3) create an EventTableEntry object out of the FullDetectedEvent, which adds the Messages and other settings
		local stub = self:FixOld_31205_CreateDEStub(oldkey)
		
		-- some spells are clearly in the spellbook (UNIT_SPELLCAST_SENT) 
		-- while others must be items (UNIT_SPELLCAST_SENT)  or procs (SPELL_AURA_APPLIED_BUFF)
		if stub.type ~= "BOTH" then
			-- stub.type is known
			local de = self:FixOld_31205_CreateDetectedEvent(stub)
			self:FixOld_31205_CreateEventTableEntry(oldkey, OldSpellTableEntry, de)
		else
			-- stub.type is unknown, record both
			self:Print("Unsure how to update: "..tostring(oldkey))--TODOFUTURE: change this to a prompt
			
			stub.type = "UNIT_SPELLCAST_SENT"
			local de = self:FixOld_31205_CreateDetectedEvent(stub)
			self:FixOld_31205_CreateEventTableEntry(oldkey, OldSpellTableEntry, de)

			stub.type = "SPELL_AURA_APPLIED_BUFF"
			de = self:FixOld_31205_CreateDetectedEvent(stub)
			self:FixOld_31205_CreateEventTableEntry(oldkey, OldSpellTableEntry, de)
		end
	end
	
	-- we can destroy the old SpellTable now.  We won't be using it anymore
	SpeakinSpellSavedData.SpellTable = nil
end


function SpeakinSpell:FixOld_31206_UpdateEventKeys()
	local funcname = "FixOld_31206_UpdateEventKeys"
	
	--changing SpeakinSpellSavedData.EventTable causes a problem while iterating on it
	--so build a new table and use it to replace the old table
	local newkeys = {}
	
	if not SpeakinSpellSavedData.EventTable then
		self:ErrorMsg(funcname, "SpeakinSpellSavedData.EventTable is nil, aborting update to 3.1.2.06")
		return
	end
	
	for oldkey,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		local de = EventTableEntry.DetectedEvent
		-- DO NOT CHANGE 3.1.2.06 format is intentional for upgrade chain
		de.key = de.type..de.name..de.rank -- DO NOT CHANGE 3.1.2.06 format
		newkeys[de.key] = EventTableEntry
	end
	
	SpeakinSpellSavedData.EventTable = self:CopyTable(newkeys)
end


function SpeakinSpell:FixOld_31207_UpdateEventKeys()
	--changing SpeakinSpellSavedData.EventTable causes a problem while iterating on it
	--so build a new table and use it to replace the old table
	local newkeys = {}
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		local de = EventTableEntry.DetectedEvent
		de.key = string.gsub( string.upper( tostring(de.type)..tostring(de.name)..tostring(de.rank)), " ", "_" ) -- DO NOT CHANGE 3.1.2.07 format
		newkeys[de.key] = EventTableEntry
	end
	SpeakinSpellSavedData.EventTable = self:CopyTable( newkeys )
end


function SpeakinSpell:FixOld_31302_AddNewEventsDetected()
	-- moved the NewEventsDetected table from RuntimeData into SpeakinSpellSavedData
	if not SpeakinSpellSavedData.NewEventsDetected then
		SpeakinSpellSavedData.NewEventsDetected = {}
	end
end


function SpeakinSpell:FixOld_31303_GetTypeFromKey(key)
	--NOTE: the order matters here
	local types = {
		[1] = "EVENT",
		[2] = "MACRO",
		[3] = "UNIT_SPELLCAST_SENT",
		[4] = "SPELL_AURA_APPLIED_BUFF_FOREIGN",
		[5] = "SPELL_AURA_APPLIED_DEBUFF_FOREIGN",
		[6] = "SPELL_AURA_APPLIED_BUFF",
		[7] = "SPELL_AURA_APPLIED_DEBUFF",
	}
	for i,type in ipairs(types) do
		if self:StartsWith(key, type) then return type end
	end
end

function SpeakinSpell:FixOld_31303_RecoverMissingTypes()
	-- I discovered that DetectedEvent.type was left undefined for many event types in the saved data
	-- thankfully, it can be recovered from the key format
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		EventTableEntry.DetectedEvent.type = self:FixOld_31303_GetTypeFromKey(key)
	end
end



function SpeakinSpell:FixOld_32006_InsertRandomFaction(name)
	if (name ~= nil) and (name ~= "") then
		tinsert(SpeakinSpellSavedData.RandomSubs["randomfaction"], name)
	end
end



function SpeakinSpell:FixOld_32006_UpgradePossessive(EventTableEntry)
	for key,msg in pairs(EventTableEntry.Messages) do
		msg = string.gsub(msg, "<focus>'s","<focus'>")
		msg = string.gsub(msg, ">'s","'s>")
		-- I think that's enough
	end
end


function SpeakinSpell:FixOld_32006_SetDefaultLeaderChannel(EventTableEntry)
	-- the Leader scenario is new in 3.2.0.06
	-- set it's default channel to match the party or raid scenarios
	if EventTableEntry.Channels.Party then
		EventTableEntry.Channels.PartyLeader = EventTableEntry.Channels.Party
	end
	if EventTableEntry.Channels.Raid then
		EventTableEntry.Channels.RaidLeader = EventTableEntry.Channels.Raid
		EventTableEntry.Channels.RaidOfficer = EventTableEntry.Channels.Raid
	end
end


function SpeakinSpell:FixOld_32006_BunchOfStuff()
	-- update EventTable data
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		-- the <pet> variable is not event-specific, but had been tracked in the DetectedEvent objects
		-- the new implementation for substitution variables might use this outdated stored value
		-- instead of the name of the currently active pet
		EventTableEntry.DetectedEvent.pet = nil
		-- update all saved messages to make use of new possessive substitutions like <player's>
		self:FixOld_32006_UpgradePossessive(EventTableEntry)
		-- update settings to use the new Leader scenario the same as SpeakinSpell would have behaved before
		self:FixOld_32006_SetDefaultLeaderChannel(EventTableEntry)
	end
	
	-- Also added <randomtaunt> and <randomfaction> to saved variables, so users can change them
	--SpeakinSpellSavedData.RandomSubs = self:CopyTable( DEFAULT_SPEECHES.RandomSubs ) -- data moved, unnecessary to import it as a patch
	--self:FixOld_32006_AddRandomFactions() -- data moved, unnecessary to import it as a patch
	
end


function SpeakinSpell:FixOld_32202_UpgradeNewline()
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		for i,Speech in pairs( EventTableEntry.Messages ) do
			Speech = string.gsub( Speech,   "<newline>",  "\n" )
			Speech = string.gsub( Speech, L["<newline>"], "\n" )
			EventTableEntry.Messages[i] = Speech
		end
	end
end



function SpeakinSpell:ApplyPatches()
	-----------------------------------------------------------------------------------
	-- Define Patches by version
	
	local Patches = {
		
		{ Version = "3.1.1.02", 
			-- Added this global toggle in 3.1.1.02, before this, it behaved as if always on
			func = function() SpeakinSpellSavedData.EnableAllMessages = true end 
		},
		
		{ Version = "3.1.2.05", 
			-- redefined the SpellTable as EventTable with a new key system and data format
			func = function() self:FixOld_31205_ImportSpellTable() end
		},
		
		{ Version = "3.1.2.06", 
			-- changed event keys so they sort by event type instead of spell name
			func = function() self:FixOld_31206_UpdateEventKeys() end
		},
		
		{ Version = "3.1.2.07",
			-- changed event keys to uppercase, with underscores replacing spaces
			func = function() self:FixOld_31207_UpdateEventKeys() end
		},

		{ Version = "3.1.3.02",
			-- moved the NewEventsDetected table from RuntimeData into SpeakinSpellSavedData
			func = function() self:FixOld_31302_AddNewEventsDetected() end
		},
		
		{ Version = "3.1.3.03",
			-- I discovered that DetectedEvent.type was left undefined for many event types in the saved data
			-- thankfully, it can be recovered from the key format
			func = function() self:FixOld_31303_RecoverMissingTypes() end
		},
		
		{ Version = "3.2.0.06",
			-- This version made several changes to the saved data
			func = function() self:FixOld_32006_BunchOfStuff() end
		},

		{ Version = "3.2.2.02",
			-- This version enhanced <newline> to allow storing "\n" in the data and GUI
			func = function() self:FixOld_32202_UpgradeNewline() end
		},
	}
	
	-----------------------------------------------------------------------------------
	-- Apply all patches

	for i, Patch in ipairs(Patches) do
		if SpeakinSpellSavedData.Version < Patch.Version then
			self:Print( string.format("Updating saved data %s -> %s", SpeakinSpellSavedData.Version, Patch.Version) )
			Patch.func() -- apply the patch defined above
			SpeakinSpellSavedData.Version = Patch.Version -- patch has been applied
		end
	end

	-----------------------------------------------------------------------------------
	-- saved data is now up to date with the current version
	SpeakinSpellSavedData.Version = SpeakinSpell.CURRENT_VERSION
end
	
